/***
	83lnk.c
	Copyleft 1997 Anthony Tong (antoine@eci.com)
	This program is placed under the GNU Licensing Agreement.

	Reads .ihx files, converting it to a 83p file.

	Note: Error checking is sparse.

BE SURE TO BACKUP YOUR CALCULATOR'S MEMORY BEFORE YOU RUN ANY COMPILED
PROGRAMS. YOU MAY HAVE TO RESET YOU CALCULATOR IN CASE YOU PROGRAM CRASHES.

* I will not be held liable for any data loss or damages.

***/

#include <stdio.h>
#define	EXIT_FAILURE	1	/* Failing exit status.  */
#define	EXIT_SUCCESS	0	/* Successful exit status.  */

#define BUFFER_SIZE	35000


FILE *infile=NULL, *outfile=NULL;
char *buffer;

#ifdef DEBUG
FILE *debug=NULL;
#endif DEBUG

const char ti83_fhdr[11] = { '*','*','T','I','8','3','*','*',0x1A,0x0A,0x00 };
const char ti83_fftr[9] = { 0x3F, 0xD4, 0x3F, '0', '0', '0', '0', 0x3F, 0xD4 };

void ProcessFile( char* progname );
void Error( char *errmsg );
char GetByte( char *buf );
long GetAddr( char *buf );


int main (int argc, char *argv[]) 
{
	unsigned char* buf;
	unsigned short cs, b, i;
	
	//	Open files
	if ( argc == 4 ) {
		if ( ( infile = fopen (argv[1],"r" ) ) == NULL ) 
			Error ( "Can't open input" );
		if ( ( outfile = fopen (argv[2],"wb") ) == NULL ) 
			Error ( "Can't open output" );
	}
	else
		Error ( "83lnk v1.0 by Anthony Tong <antoine@eci.com>\n\n"
		"Usage: 83lnk infile.obj outfile.83p PROGNAME" );
#ifdef DEBUG
	if ( ( debug = fopen ( "83.obj", "w" ) ) == NULL )
		Error ( "Can't open debug file" );
#endif DEBUG

	if ( (buffer = (char * ) malloc ( BUFFER_SIZE )) == NULL )
		Error ( "Can't allocate memory" );

	ProcessFile( argv[3] );


	// NEED TO SET CHECKSUM!!

	infile=fopen( argv[2], "rb" );
	fseek( infile, 0x35, SEEK_SET );
	fread( &b, sizeof(short), 1, infile );
	fread( buffer, sizeof(char), b, infile );
	fclose ( infile );

	cs = 0x0000;
	buf = (unsigned char *) buffer;
	for ( i=0; i<b; i++ )
		cs += buf[i];

	outfile=fopen( argv[2], "ab" );
	fwrite( &cs, sizeof(short), 1, outfile);

	fclose ( outfile );
#ifdef DEBUG
	fclose ( debug );
#endif DEBUG

	return ( EXIT_SUCCESS );
}

void ProcessFile( char* progname ) {
	long addr, cur_addr, real_bytes;
	int eof;
	char bytes;
	short b;

	//	Write file header; skip over addresses & checksums
	fwrite ( ti83_fhdr, sizeof(char), 11, outfile );
	//	zero out what we're writing.. needed for DOS
	for ( b=0; b<=0x3D; b++ )
		buffer[b] = 0x00;
	fwrite ( buffer, sizeof(char), 0x3D, outfile );

	eof = (int) fgets ( buffer, 2000, infile );
	real_bytes = 0;
	bytes = GetByte ( buffer ); 
	cur_addr = 0x9327;

	//	File must start of at 0x9327;
	if ( (addr = GetAddr ( buffer )) != cur_addr )
		Error ( "Source is not TI83 specific" );

	while ( eof ) {
		fwrite ( (buffer+9), bytes*2, sizeof(char), outfile );
#ifdef DEBUG
		fwrite ( (buffer+9), bytes*2, sizeof(char), debug );
#endif DEBUG
		real_bytes += ( bytes * 2 );
		cur_addr = addr + bytes;
		eof = (int) fgets ( buffer, 2000, infile );
		bytes = GetByte ( buffer );
		addr = GetAddr ( buffer );

		// fill in bytes left by .dw
		while ( cur_addr < addr ) {
			fputc ( '0', outfile );
			fputc ( '0', outfile );
#ifdef DEBUG
			fputc ( '0', debug );
			fputc ( '0', debug );
#endif DEBUG
			real_bytes += 2;
			cur_addr ++;
		}
	}	
	fwrite ( ti83_fftr, sizeof(char), 9, outfile );
	real_bytes += 11;

	fseek ( outfile, 0x35, SEEK_SET );
	b = (short) real_bytes + 0x0F;
	fwrite ( &b, sizeof(short), 1, outfile );
	fputc( 0x0B, outfile );
	fputc( 0x00, outfile );
	b = (short) real_bytes;
	fwrite ( &b, sizeof(short), 1, outfile );
	fseek ( outfile, 0x3B, SEEK_SET );
	fputc ( 0x05, outfile );
	fprintf ( outfile, progname );

	fseek ( outfile, 0x44, SEEK_SET );
	fwrite ( &b, sizeof(short), 1, outfile );
	b = (short) real_bytes - 2;
	fwrite ( &b, sizeof(short), 1, outfile );

	fclose ( infile );
	fclose ( outfile );
}

char GetByte ( char *buf ) {
	char b;

	if ( *buf != ':' )
		Error ( "Error reading file" );
	// Conversion from ascii to hex
	b = ( buf[2] - ( ( buf[2] < 65 ) ? 48 : 55 ) )
		+ ( 16 * ( buf[1] - ( ( buf[1] < 65 ) ? 48 : 55 ) ) );

	return b;
}

long GetAddr( char *buf ) {
	long a;

	a = ( buf[6] - ( ( buf[6] < 65 ) ? 48 : 55 ) )
		+ ( 16 * ( buf[5] - ( ( buf[5] < 65 ) ? 48 : 55 ) ) )
		+ ( 256 * ( buf[4] - ( ( buf[4] < 65 ) ? 48 : 55 ) ) )
		+ ( 4096 * ( buf[3] - ( ( buf[3] < 65 ) ? 48 : 55 ) ) );
	return a;
}

void Error ( char *errmsg ) {
	fprintf ( stderr, errmsg );
	fprintf ( stderr, "\n\n" );
	exit ( EXIT_FAILURE );
}

